/**
 * @author vedder
 * @copyright 2008 changeMode design
 */
function App(){
	var self = this;  // for self-refferential goodness
	this.timer = new Timer(this); 
	// set up prefs file business
	this.prefsFile = air.File.applicationStorageDirectory;
	this.prefsFile = this.prefsFile.resolvePath("timerPreferences.xml");
	/// set up sound business
	this.defaultAlertSound = new air.Sound(new air.URLRequest('sound/alert.mp3'));
	this.customSoundFile = null;
	this.soundReference = null; // will hold the ref to the sound we have to use to stop it playing.
	// create jQuery objects of DOM stuff we interact with a bunch
	this.closeButton = $('#closeButton');
	this.growHandle = $('#growHandle');
	this.onTopButton = $('#onTopButton');
	this.centerSection = $('#centerSection');
	this.startStopButton = $('#startStopButton');
	this.clearButton = $('#clearButton');
	this.hoursDisplay = $('#hoursDisplay');
	this.minutesDisplay = $('#minutesDisplay');
	this.secondsDisplay = $('#secondsDisplay');
	this.soundChooser = $('#soundChooser');
	this.customLabel = $('#customLabel');
	this.customSoundButton = $('#customSoundButton');
	this.soundRemoveButton = $('#soundRemoveButton');
	this.defaultSoundButton = $('#defaultSoundButton');
	this.prefsSoundFlag = false; // so the sound UI method does not over-ride user prefs
	
	///////////////////////////////
	// priviledged functions (ones that get called by timers and so on that need 'self')
	
	/**
 * Handles the native window close event and cancels it, allowing us to execute any code we want first 
 * like saving prefs or anything
 * 
 * @param {Object} event 
 * @return void;
 */
	this.windowClosingHandler = function(event){
		event.preventDefault();
		nativeWindow.removeEventListener("closing", self.windowClosingHandler);
		// our code to execute

		// now exit.
		air.NativeApplication.nativeApplication.exit();
	}
	
	this.applicationExit = function (){ 
    var exitingEvent = new air.Event(air.Event.EXITING, false, true); 
    air.NativeApplication.nativeApplication.dispatchEvent(exitingEvent); 
    if (!exitingEvent.isDefaultPrevented()) { 
        air.NativeApplication.nativeApplication.exit(); 
    } 
	}
	
	this.appShutdownHandler = function(){
				self.writePreferenceFile();
	}
	/*
	 * 
	 * 
	 * You're doing a couple of strange things here:

- Calling preventDefault on the CLOSING event. That prevents the window from being closed. Since you're supposedly trying to exit, why keep the window open? I believe this is interrupting the shutdown sequence.

- If you want to do something on exit, you should be listening for EXITING, not CLOSING.

- If you want a button that quits the app, it should call exit(), not dispatch a CLOSING event.

- If you want a button that closes a window, it should call close(), not dispatch closing().

Finally, the application menu item and Cmd-Q work correctly by default. Adding event handlers, etc. for these events is only necessary if you want to override the default behavior.
	 */

	
	/**
	 * Reveals the center content area and resized the application window
	 * @param {event} event.click The mouseclick event from the open handle
	 * @return {void}
	 */
	this.showCenter=  function (event){
		event.stopPropagation();
		window.resizeBy(0,self.centerSection.height());
		self.centerSection.slideDown(500);
		self.growHandle.unbind('click'); 
		self.growHandle.click(self.hideCenter); 
	}
	
	/**
	 * Animates hiding of the center area and resizes the application window down.
	 * @return {void} 
	 */
	this.hideCenter = function (){
		self.centerSection.slideUp(500,function(){	window.resizeBy(0,-1*self.centerSection.height()); });
		self.growHandle.unbind('click'); 
		self.growHandle.click(self.showCenter)
	}
	
	/**
	 * Toggles if the application window is always on top or not
	 * @return {void}
	 */
	this.toggleOnTop =  function (){
		if (nativeWindow.alwaysInFront) {
			nativeWindow.alwaysInFront = false;
			self.onTopButton.attr('src',"img/onTopButton_inactive.png");;
		}else{
			nativeWindow.alwaysInFront = true;
			self.onTopButton.attr('src',"img/onTopButton_active.png");
		}
	}
	
	/**
	 * Nulls out our custom sound fiel and updates GUI 
	 * return {void}
	 */
	this.removeCustomSound = function(){
		self.customSoundFile = null;
		self.soundRemoveButton.fadeOut("slow");
		self.customLabel.text("Choose...").addClass('greyOut');
		self.customSoundButton.click(self.chooseSoundFile);
		self.defaultSoundButton.attr("checked",true);
	}
	
	/**
	 * Handles the dropping of files where we listen for it and 
	 * if one MP3 files is dropped, we set as custom sound file
	 * 
	 * @param {event.drop} event
	 * @return {void}
	 */
	this.dropHandler = function (event){
		event.preventDefault();
		var fileList = event.dataTransfer.getData("application/x-vnd.adobe.air.file-list");
		if(fileList.length !== 1){
			alert("Dude, I can only handle one file");
		}else{
			var myFile = new air.File();
			var draggedFile = fileList[0];
			if(draggedFile.extension == 'mp3'){
				self.prefsSoundFlag = false;
				self.loadCustomSound(draggedFile.url);
			}else{
				alert("Sorry, only .mp3 files are suported at this time.");
			}
		}
		self.soundChooser.removeClass('draggedOver');
	}
	
	/**
	 * Handles both dragenter and dragover events to hilight drop zone and to 
	 * use our behavior instead of default
	 * 
	 * @param {event.drag} event
	 * @return {void}
	 */
	this.dragInHandler = function (event){
		event.dataTransfer.dropEffect = "link";
		event.preventDefault();
		self.soundChooser.addClass('draggedOver');
	}
	
	/**
	 * Called when user has chosen a sound form the fiel browser to set sound reference and update GUI
	 * 
	 * @param {event.select} event
	 * @return {void}
	 */
	this.fileSelectHandler = function (event){
		self.prefsSoundFlag = false;
		self.loadCustomSound(event.target.url);
	}
	
	/**
	 * opens file system browser for user to choose sound file.
	 * 
	 * @return {void}
	 */
	this.chooseSoundFile = function (){
		var theFile = new air.File();
		var filter = new air.FileFilter("MP3","*.mp3");
		theFile.browseForOpen("Open",new window.runtime.Array(filter));
		theFile.addEventListener(air.Event.SELECT, self.fileSelectHandler);
	}
	
	/**
	 * Plays and sets up looping for sound file
	 * 
	 * @param {Sound} snd
	 * @return {void}
	 */
	this.playSound = function (snd){
 		self.soundReference = snd.play();
		self.soundReference.addEventListener(air.Event.SOUND_COMPLETE, function(){ self.playSound(snd); }); // loopy goodness
	}
	
	/**
	 * stops playback of current sound.
	 * 
	 * @return {void}
	 */
	this.stopSound = function (){
		if (self.soundReference !== null) {
			self.soundReference.stop();
		}
	}
	
	/**
 * Updates GUI when we have a new custom sound file
 * @return {void}
 */
	this.updateSoundUI = function(){
		self.customLabel.text(self.getFileName(self.customSoundFile.url)).removeClass('greyOut');
		self.customSoundButton.unbind('click')
		if (!self.prefsSoundFlag) {
			self.customSoundButton.attr("checked", true);
			self.prefsSoundFlag = false;
		}
		self.soundRemoveButton.fadeIn("slow");
	}

		
	// end of priviledged functions
	////////////////////////////////
	
	////////////////////////
	// Look kids!! Event handler fun! Makes app do stuff! Order now!!	
//	window.nativeWindow.addEventListener(air.Event.CLOSING, this.windowClosingHandler);
	air.NativeApplication.nativeApplication.addEventListener( air.Event.EXITING, this.appShutdownHandler);

//	this.closeButton.click(function(e){ 
		//var doClose = new air.Event(air.Event.CLOSING, true, true);
		//window.nativeWindow.dispatchEvent(doClose);
//	});

	$('body').mousedown(function(){nativeWindow.startMove();}).bind('contextmenu',function(e){return false;}); // drag by anything, disable right click menu
	// remove ability to drag by things we don't want to be dragged by, and from selecting text.
	$('.noDrag').mousedown(function(event){return false}); 
	this.growHandle.click(this.showCenter);
	this.onTopButton.click(this.toggleOnTop);
	this.clearButton.click(this.timer.clearTimer);
	this.startStopButton.click(this.timer.start);
	// TODO: Figure out why these fail miserably as a privledged functions so we don't have to pass a reference.
	this.secondsDisplay.bind('click',{val:1,app:this},this.timerClickHandler);
	this.minutesDisplay.bind('click',{val:60,app:this},this.timerClickHandler);
	this.hoursDisplay.bind('click',{val:3600,app:this},this.timerClickHandler);
	
	this.soundChooser.bind('dragover',this.dragInHandler);
	this.soundChooser.bind('dragenter',this.dragInHandler);
	this.soundChooser.bind('drop',this.dropHandler);
	
	this.customSoundButton.click(this.chooseSoundFile);
	this.soundRemoveButton.click(this.removeCustomSound);
	
	this.closeButton.click(this.applicationExit);
	// end of event handlers
	/////////////////////////////
	
	///////////////////////
	// read prefs file and parse it!
	if(this.prefsFile.exists){
		this.readPreferenceFile();
	}else{
		// write a default pres file
		this.writePreferenceFile();
	}
	
}

/**
 * Updates UI when timer has fired a start event (it has started)
 * 
 * @param {timerStartEvent} event
 * @return {Boolean}
 */
App.prototype.timerStartHandler = function (event){
	this.clearButton.unbind('click').addClass('greyOut');
	this.startStopButton.text("STOP").unbind('click').click(this.timer.stop);
	return false;
}

/**
 * Updates UI when Timer has fired a stop event
 * @param {timerStopEvent} event
 * @return {Boolean}
 */
App.prototype.timerStopHandler = function (event){
	this.clearButton.click(this.timer.clearTimer).removeClass('greyOut');
	this.startStopButton.text("START").unbind('click').click(this.timer.start).removeClass('greyOut');
	return false;
}

/** 
 * Updates UI when Timer has fired a clear event
 * @param {timerClearEvent} event
 * @return {Boolean}
 */
App.prototype.timerClearHandler = function (event){
	if (this.timer.isRunning()) {		this.timer.stop();	}
	$('.timerDisplay').removeClass('negativeValue');
	this.stopSound();
	return false;
}


/**
 * Handles event from timer when it has hit zero
 * @param {timerAlertEvent} event
 * @return {Boolean}
 */
App.prototype.timerAlertHandler = function (event){
		$('.timerDisplay').addClass('negativeValue');
		var useSound = (this.customSoundButton.attr("checked") === true) ? this.customSoundFile  : this.defaultAlertSound;
		this.playSound( useSound);
		this.startStopButton.unbind('click').addClass('greyOut');
		this.clearButton.click(this.timer.clearTimer).removeClass('greyOut');
		// ounce for OS X folks, maybe if it works
		if(air.NativeApplication.supportsDockIcon){
     var dockIcon = air.NativeApplication.nativeApplication.icon;
     dockIcon.bounce(air.NotificationType.CRITICAL);
 		}
		return false;
}

/**
 * Handles event from Timer when it is updated.
 * @param {timerUpdateEvent} event 
 * @param {Number} val
 * @return {Boolean}
 */
App.prototype.timerUpdateHandler =  function (event,val){
	var val = (val < 0) ? val*-1 : val;
	this.updateDisplay(val);
	return false;
}

/**
 * Increments Timer value based on user click. Modifies value based on modifer keys: Shift = subtract, 
 * Control = 10 units. Units are passed as a data object on the event.
 * @param {Event.Click} event
 * @return {void}
 */
App.prototype.timerClickHandler = function (event){
	var amount = (event.shiftKey) ? event.data.val * -1 : event.data.val;
	if(event.ctrlKey) amount = amount *10;
	if(event.data.app.timer.value >= 0 && event.data.app.timer.value + amount > 0) event.data.app.timer.increment(amount);
	// technicaly we shouldn't let people go over 99 hours, 59 mins and 59 sec, as that will display
	// only the last two digits of the hours, but if you really want to time that long, you deserve dispaly problems.
}

/**
 * Updates the timer display fields using the given value
 * @param {Number} val
 * @return {void}
 */
App.prototype.updateDisplay = function(val){
		var hr = Math.floor(val / 3600);
		this.hoursDisplay.text(this.formatNums(hr));
		var min = Math.floor((val - 3600 * hr) / 60);
		this.minutesDisplay.text( this.formatNums(min));
		this.secondsDisplay.text( this.formatNums( Math.floor(val - 3600 * hr - min * 60)) );
	}
	
/**
 * return url decoded name of a file without extension form file URL. Will truncate with elipsis if over 17 in length
 * 
 * @param {String} theFilePath filepath URL of sound file
 * @return {String}
 */	
App.prototype.getFileName = function (theFilePath)	{
	var tmp = theFilePath.split('/'); // air.File.separator if we were doing nativePath
	var longName = unescape(tmp[tmp.length-1]);
	var shortName = longName.substring(0,longName.lastIndexOf('.'));
	if(shortName.length > 17){ shortName = shortName.substring(0,14) + "...";}
	return shortName;
}

/**
 * synchronously writes XML preferences file with data from currnent state of app
 * 
 * @return {void}
 */
App.prototype.writePreferenceFile = function(){
	var lf = air.File.lineEnding;
	var useCustomSound = (this.customSoundButton.attr("checked") === true) ? 'yes' : 'no';
	var customSoundURL = (this.customSoundFile === null) ? 'none' : this.customSoundFile.url;
	var prefXML = "<?xml version='1.0' encoding='utf-8'?>" + lf + "<preferences>"+lf;
		prefXML += "<windowPosition y='"+ window.screenTop.toString() +"' x='"+ window.screenLeft.toString() +"' foreground='"+ nativeWindow.alwaysInFront +"' />"+ lf;
		prefXML += "<customSound enabled='" + useCustomSound +"'>";
		prefXML += customSoundURL;
		prefXML += "</customSound>"+lf;
		prefXML += "</preferences>";
		////////////////////////////
		
	var fs = new air.FileStream();
	fs.open(this.prefsFile, air.FileMode.WRITE);
	fs.writeUTFBytes(prefXML);
	fs.close();
}

/**
 * Synchronously reads existing prefs file and upadtes GUI with information it contains.
 * @return {void}
 */
App.prototype.readPreferenceFile = function(){
	var fs = new air.FileStream();
	fs.open(this.prefsFile,air.FileMode.READ);
	var prefXML = fs.readUTFBytes(fs.bytesAvailable);
	fs.close();

	var myParser = new DOMParser();
	prefXML = myParser.parseFromString(prefXML,"text/xml");
	var winPos = prefXML.getElementsByTagName("windowPosition")[0];
	var customSoundInfo = prefXML.getElementsByTagName("customSound")[0];
	if(customSoundInfo.getAttribute('enabled') === 'no'){ this.prefsSoundFlag = true; }
	if(customSoundInfo.firstChild.nodeValue !== 'none'){ this.loadCustomSound(customSoundInfo.firstChild.nodeValue); }
	if (winPos.getAttribute('foreground') == 'true') { this.toggleOnTop(); }
	window.moveTo(winPos.getAttribute('x'),winPos.getAttribute('y'));
	
}

/**
 * Attmpts to make Sound object form file URL
 * @param {String} fileURL
 * @return {void}
 */
App.prototype.loadCustomSound = function (fileURL){
	this.customSoundFile = new air.Sound(new air.URLRequest(fileURL));
	// in theory this only happens when a file in the prefs.xml has been moved or deleted.
	this.customSoundFile.addEventListener(air.IOErrorEvent.IO_ERROR, this.removeCustomSound);
	this.customSoundFile.addEventListener(air.Event.COMPLETE,this.updateSoundUI);
}


/**
 * Takes a numeric value and pads with a zero in front if it is less than 10
 * @param {Number} val
 * @return {String}
 */
App.prototype.formatNums = function(val){
	return val > 9 ? String(val) : "0"+ String(val);
}